#pragma once

#include <chrono>
#include <memory>
#include <optional>
#include <session/config.hpp>

#include "base.hpp"
#include "namespaces.hpp"
#include "notify.hpp"
#include "profile_pic.hpp"
#include "theme.hpp"

namespace session::config {

using namespace std::literals;

/// keys used in this config, either currently or in the past (so that we don't reuse):
///
/// n - user profile name
/// p - user profile url
/// q - user profile decryption key (binary)
/// + - the priority value for the "Note to Self" pseudo-conversation (higher = higher in the
///     conversation list).  Omitted when 0.  -1 means hidden.
/// e - the expiry timer (in seconds) for the "Note to Self" pseudo-conversation.  Omitted when 0.
/// M - set to 1 if blinded message request retrieval is enabled, 0 if retrieval is *disabled*, and
///     omitted if the setting has not been explicitly set (or has been explicitly cleared for some
///     reason).

class UserProfile : public ConfigBase {

  private:
    notify_content notification_content_setting = notify_content::defaulted;
    notify_sound notification_sound_setting = notify_sound::defaulted;
    theme theme_setting = theme::defaulted;
    theme_primary_color theme_primary_color_setting = theme_primary_color::defaulted;
    std::map<std::string, uint16_t> local_settings;

  public:
    // No default constructor
    UserProfile() = delete;

    /// API: user_profile/UserProfile::UserProfile
    ///
    /// Constructs a user profile from existing data (stored from `dump()`) and the user's secret
    /// key for generating the data encryption key.  To construct a blank profile (i.e. with no
    /// pre-existing dumped data to load) pass `std::nullopt` as the second argument.
    ///
    /// Inputs:
    /// - `ed25519_secretkey` -- contains the libsodium secret key used to encrypt/decrypt the
    /// data when pushing/pulling from the swarm.  This can either be the full 64-byte value (which
    /// is technically the 32-byte seed followed by the 32-byte pubkey), or just the 32-byte seed of
    /// the secret key.
    /// - `dumped` -- either `std::nullopt` to construct a new, empty object; or binary state data
    /// that was previously dumped from an instance of this class by calling `dump()`.
    ///
    /// Outputs:
    /// - `UserProfile` - Constructor
    UserProfile(
            std::span<const unsigned char> ed25519_secretkey,
            std::optional<std::span<const unsigned char>> dumped);

    /// API: user_profile/UserProfile::storage_namespace
    ///
    /// Returns the UserProfile namespace. Is constant, will always return 2
    ///
    /// Inputs: None
    ///
    /// Outputs:
    /// - `Namespace` - Will return 2
    Namespace storage_namespace() const override { return Namespace::UserProfile; }

    /// API: user_profile/UserProfile::encryption_domain
    ///
    /// Returns the domain. Is constant, will always return "UserProfile"
    ///
    /// Inputs: None
    ///
    /// Outputs:
    /// - `const char*` - Will return "UserProfile"
    const char* encryption_domain() const override { return "UserProfile"; }

    /// API: user_profile/UserProfile::get_name
    ///
    /// Returns the user profile name, or std::nullopt if there is no profile name set.
    ///
    /// Inputs: None
    ///
    /// Outputs:
    /// - `std::optional<std::string>` - Returns the user profile name if it exists
    std::optional<std::string_view> get_name() const;

    /// API: user_profile/UserProfile::set_name
    ///
    /// Sets the user profile name; if given an empty string then the name is removed.
    ///
    /// Inputs:
    /// - `new_name` -- The name to be put into the user profile
    void set_name(std::string_view new_name);

    /// API: user_profile/UserProfile::set_name_truncated
    ///
    /// Sets the user profile name; if given an empty string then the name is removed. Same as the
    /// `set_name` function but truncates the name if it's too long.
    ///
    /// Inputs:
    /// - `new_name` -- The name to be put into the user profile
    void set_name_truncated(std::string new_name);

    /// API: user_profile/UserProfile::get_profile_pic
    ///
    /// Gets the user's current profile pic URL and decryption key.  The returned object will
    /// evaluate as false if the URL and/or key are not set.
    ///
    /// Inputs: None
    ///
    /// Outputs:
    /// - `profile_pic` - Returns the profile pic
    profile_pic get_profile_pic() const;

    /// API: user_profile/UserProfile::set_profile_pic
    ///
    /// Sets the user's current profile pic to a new URL and decryption key.  Clears both if either
    /// one is empty.
    ///
    /// Declaration:
    /// ```cpp
    /// void set_profile_pic(std::string_view url, std::span<const unsigned char> key);
    /// void set_profile_pic(profile_pic pic);
    /// ```
    ///
    /// Inputs:
    /// - First function:
    ///    - `url` -- URL pointing to the profile pic
    ///    - `key` -- Decryption key
    /// - Second function:
    ///    - `pic` -- Profile pic object
    void set_profile_pic(std::string_view url, std::span<const unsigned char> key);
    void set_profile_pic(profile_pic pic);

    /// API: user_profile/UserProfile::get_nts_priority
    ///
    /// Gets the Note-to-self conversation priority.  Negative means hidden; 0 means unpinned;
    /// higher means higher priority (i.e. hidden in the convo list).
    ///
    /// Inputs: None
    ///
    /// Outputs:
    /// - `int` - Returns a numeric representing prioritity
    int get_nts_priority() const;

    /// API: user_profile/UserProfile::set_nts_priority
    ///
    /// Sets the Note-to-self conversation priority. -1 for hidden, 0 for unpinned, higher for
    /// pinned higher.
    ///
    /// Inputs:
    /// - `priority` -- Numeric representing priority
    void set_nts_priority(int priority);

    /// API: user_profile/UserProfile::get_nts_expiry
    ///
    /// Returns the current Note-to-self message expiry timer, if set, or std::nullopt if there is
    /// no current expiry timer set.
    ///
    /// Inputs: None
    ///
    /// Outputs:
    /// - `std::optional<std::chrono::seconds>` - Returns the timestamp representing the message
    /// expiry timer if the timer is set
    std::optional<std::chrono::seconds> get_nts_expiry() const;

    /// API: user_profile/UserProfile::set_nts_expiry
    ///
    /// Sets the Note-to-self message expiry timer.  Call without arguments (or pass a zero time) to
    /// disable the expiry timer.
    ///
    /// Inputs:
    /// - `timer` -- Default to 0 seconds, will set the expiry timer
    void set_nts_expiry(std::chrono::seconds timer = 0s);

    /// API: user_profile/UserProfile::get_blinded_msgreqs
    ///
    /// Accesses whether or not blinded message requests are enabled for the client.  Can have three
    /// values:
    ///
    /// - std::nullopt -- the value has not been given an explicit value so the client should use
    ///   its default.
    /// - true -- the value is explicitly enabled (i.e. user wants blinded message requests)
    /// - false -- the value is explicitly disabled (i.e. user disabled blinded message requests)
    ///
    /// Inputs: None
    ///
    /// Outputs:
    /// - `std::optional<bool>` - true/false if blinded message requests are enabled or disabled;
    ///   `std::nullopt` if the option has not been set either way.
    std::optional<bool> get_blinded_msgreqs() const;

    /// API: user_profile/UserProfile::set_blinded_msgreqs
    ///
    /// Sets whether blinded message requests (i.e. from SOGS servers you are connected to) should
    /// be enabled or not.  This is typically invoked with either `true` or `false`, but can also be
    /// called with `std::nullopt` to explicitly clear the value.
    ///
    /// Inputs:
    /// - `enabled` -- true if blinded message requests should be retrieved, false if they should
    ///   not, and `std::nullopt` to drop the setting from the config (and thus use the client's
    ///   default).
    void set_blinded_msgreqs(std::optional<bool> enabled);

    /// API: user_profile/UserProfile::get_notification_content
    ///
    /// Returns the locally stored setting indicating what notification content should be displayed.
    ///
    /// Inputs: None
    ///
    /// Outputs:
    /// - `notify_content` -- enum indicating the content that should be shown within a
    /// notification.
    notify_content get_notification_content() const;

    /// API: user_profile/UserProfile::set_notification_content
    ///
    /// Sets the locally stored setting indicating what notification content should be displayed.
    ///
    /// Inputs:
    /// - `value` -- Updated notification content setting
    void set_notification_content(notify_content value);

    /// API: user_profile/UserProfile::get_notification_sound
    ///
    /// Returns the locally stored setting indicating which sound should play when receiving a
    /// notification.
    ///
    /// Inputs: None
    ///
    /// Outputs:
    /// - `notify_sound` -- enum indicating the sound that should be played when receiving a
    /// notification.
    notify_sound get_notification_sound() const;

    /// API: user_profile/UserProfile::set_notification_sound
    ///
    /// Sets the locally stored setting indicating which sound should be played when
    /// receiving receiving a notification.
    ///
    /// Inputs:
    /// - `value` -- Updated notification sound setting
    void set_notification_sound(notify_sound value);

    /// API: user_profile/UserProfile::get_theme
    ///
    /// Returns the locally stored setting indicating which theme the client should use.
    ///
    /// Inputs: None
    ///
    /// Outputs:
    /// - `theme` -- enum indicating which theme the client should use.
    theme get_theme() const;

    /// API: user_profile/UserProfile::set_theme
    ///
    /// Sets the locally stored setting indicating which theme the client should use.
    ///
    /// Inputs:
    /// - `value` -- Updated theme setting
    void set_theme(theme value);

    /// API: user_profile/UserProfile::get_theme_primary_color
    ///
    /// Returns the locally stored setting indicating which primary color the client should use.
    ///
    /// Inputs: None
    ///
    /// Outputs:
    /// - `theme_primary_color` -- enum indicating which primary color the client should use.
    theme_primary_color get_theme_primary_color() const;

    /// API: user_profile/UserProfile::set_theme_primary_color
    ///
    /// Sets the locally stored setting indicating which primary color the client should use.
    ///
    /// Inputs:
    /// - `value` -- Updated primary color setting
    void set_theme_primary_color(theme_primary_color value);

    /// API: user_profile/UserProfile::get_local_setting
    ///
    /// Accesses the locally stored setting for the provided key.  Can have three
    /// values:
    ///
    /// - std::nullopt -- the value has not been given an explicit value so the client should use
    ///   its default.
    /// - true -- the value is explicitly enabled
    /// - false -- the value is explicitly disabled
    ///
    /// Inputs:
    /// - `key` -- key that a setting was previously stored against.
    ///
    /// Outputs:
    /// - `std::optional<bool>` - true/false if the value has been set;
    ///   `std::nullopt` if the value has not been set.
    std::optional<bool> get_local_setting(std::string key) const;

    /// API: user_profile/UserProfile::set_local_setting
    ///
    /// Sets the locally stored setting.  This is typically invoked with either `true` or `false,
    /// but can also be called with `std::nullopt` to explicitly clear the value.
    ///
    /// Inputs:
    /// - `key` -- key that a setting was previously stored against.
    /// - `enabled` -- value that should be stored locally against the key, or `std::nullopt` to
    /// drop the setting from the local storage (and thus use the client's default).
    void set_local_setting(std::string key, std::optional<bool> enabled);

    bool accepts_protobuf() const override { return true; }

  protected:
    void extra_data(oxenc::bt_dict_producer&& extra) const override;
    void load_extra_data(oxenc::bt_dict_consumer&& extra) override;
};

}  // namespace session::config
